//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"encoding/json"
	"net/http"
	"strings"

	"github.com/gorilla/mux"
)

var auth = &Auth{}

type Auth struct {
	sessions *SessionManager
	oidc     *OIDCBackend
	config   *Config
}

func Init(c *Config) (err error) {
	if c == nil {
		// authentication is disabled
		return
	}
	auth.config = c

	if auth.sessions, err = NewSessionManager(); err != nil {
		return
	}

	if c.OIDC != nil {
		if auth.oidc, err = NewOIDCBackend(c.OIDC); err != nil {
			return
		}
	}
	return
}

func login(w http.ResponseWriter, r *http.Request) {
	backend := r.URL.Query().Get("backend")
	switch strings.ToLower(backend) {
	case "oidc":
		if auth.oidc == nil {
			resp := HTTPResponse{Error: "OIDC authentication is not configured"}
			sendHTTPResponse(w, http.StatusBadRequest, resp)
			return
		}
		auth.oidc.HandleLogin(w, r)
	case "":
		resp := HTTPResponse{Error: "default/fallback authentication backend has not been implemented yet"}
		sendHTTPResponse(w, http.StatusNotImplemented, resp)
	default:
		resp := HTTPResponse{Error: "invalid authentication backend: " + backend}
		sendHTTPResponse(w, http.StatusBadRequest, resp)
	}
}

func logout(w http.ResponseWriter, r *http.Request) {
	sid, s := getSessionFromCookie(r)
	invalidateSessionCookie(w)
	if s == nil {
		sendHTTPResponseInvalidSession(w)
		return
	}
	auth.sessions.remove(sid)
	sendHTTPResponse(w, http.StatusOK, HTTPResponse{Message: "you are now logged out"})
}

func whoami(w http.ResponseWriter, r *http.Request) {
	_, s := getSessionFromCookie(r)
	if s == nil {
		sendHTTPResponseInvalidSession(w)
		return
	}

	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)
	json.NewEncoder(w).Encode(s)
}

type AuthBackendInfo struct {
	Name        string
	Description string
}

func listBackends(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)

	backends := []AuthBackendInfo{}
	if auth.oidc != nil {
		backend := AuthBackendInfo{Name: "oidc"}
		backend.Description = "OpenID Connect using Identity Provider: " + auth.config.OIDC.IssuerURL
		backends = append(backends, backend)
	}

	json.NewEncoder(w).Encode(backends)
}

func disabled(w http.ResponseWriter, r *http.Request) {
	sendHTTPResponse(w, http.StatusBadRequest, HTTPResponse{Error: "authentication is disabled"})
}

func InstallHTTPHandler(r *mux.Router) {
	if auth.sessions == nil {
		r.PathPrefix("/").HandlerFunc(disabled)
		return
	}

	r.HandleFunc("/login", login)
	r.HandleFunc("/logout", logout)
	r.HandleFunc("/whoami", whoami)
	r.HandleFunc("/backends", listBackends)
	if auth.oidc != nil {
		r.Handle("/oidc/callback", auth.oidc.CallbackHandler())
	}
}

func Middleware(next http.Handler) http.Handler {
	if auth.sessions == nil {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			next.ServeHTTP(w, attachSessionToRequest(r, anonAllowAll))
		})
	}

	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		_, s := getSessionFromCookie(r)
		if s == nil {
			sendHTTPResponseInvalidSession(w)
			return
		}
		next.ServeHTTP(w, attachSessionToRequest(r, s))
	})
}
