/*
 *  tank
 *
 *  Import and Playlist Daemon for autoradio project
 *
 *
 *  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
 *
 *  This file is part of tank.
 *
 *  tank is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  tank is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with tank. If not, see <http://www.gnu.org/licenses/>.
 */

'use strict';

/********************************* utils *********************************/

Number.prototype.pad = function(size) {
  var s = String(this);
  while (s.length < (size || 2)) {s = "0" + s;}
  return s;
};

function format_datetime(d) {
  if(Object.prototype.toString.call(d) !== '[object Date]') {
    return '-';
  }
  if (d.toString() === 'Invalid Date') {
    return '-';
  }
  var datetimestr = Number(d.getDate()).pad(2);
  datetimestr += '.' + Number(d.getMonth() + 1).pad(2);
  datetimestr += '.' + d.getFullYear();
  datetimestr += ' ' + Number(d.getHours()).pad(2);
  datetimestr += ':' + Number(d.getMinutes()).pad(2);
  datetimestr += ':' + Number(d.getSeconds()).pad(2);
  return datetimestr;
}

function format_duration(time) {
  if(isNaN(time)) return '-';

    // duration is in ns -> first convert to ms
  time = Number(Math.floor(time / 1000000));
  var h = Number(Math.floor(time / 3600000));
  time %= 3600000;
  var m = Number(Math.floor(time / 60000));
  time %= 60000;
  var s = Number(Math.floor(time / 1000));
  var hs = Number(Math.floor((time % 1000)/100));

  return h + ':' + m.pad(2) + ':' + s.pad(2) + '.' + hs;
}

/********************************* main *********************************/

var main = main || {};

/***************** models *****************/

/****** authInfo ******/

var auth = null;

main.AuthInfo = function() {
  this.$this = $(this);
  this.logged_in = false;
  this.session = {};
}

main.AuthInfo.prototype.fetch = function() {
  var self = this;

  $.getJSON( "/auth/whoami", function(data) {
    self.session = data;
    self.logged_in = true;
    self.$this.trigger('update');
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      alert("connection problem");
    } else {
      self.logged_in = false;
      self.session = {};
      self.$this.trigger('update');
    }
  });
};

main.AuthInfo.prototype.login = function() {
  var self = this;

  $.getJSON( "/auth/login?backend=oidc", function(data) {
    self.fetch();
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      alert("connection problem");
    } else {
      alert(error);
    }
  });
};

main.AuthInfo.prototype.logout = function() {
  var self = this;

  $.getJSON( "/auth/logout", function(data) {
    self.fetch();
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      alert("connection problem");
    } else {
      alert(error);
    }
  });
};

/****** showList ******/

var shows = null;

main.ShowList = function() {
  this.$this = $(this);
  this.shows = {};
};

main.ShowList.prototype.error = function(error) {
  this.last_error = error;
  this.$this.trigger('error');
};

main.ShowList.prototype.fetch = function() {
  var self = this;
  self.last_error = null;

  $.getJSON( "/api/v1/shows", function(data) {
    self.shows = [];
    for(var key in data.results) {
      self.shows[data.results[key].name] = new main.Show(data.results[key]);
    }
    self.$this.trigger('update');
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      self.error("connection problem");
    } else {
      self.error(error);
    }
  });
};

main.ShowList.prototype.new_show = function(show_name) {
  var self = this;
  self.last_error = null;

  $.ajax({
    method: "POST",
    url: "/api/v1/shows/" + show_name,
    success: function(data) { self.fetch() },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      self.error("connection problem");
    } else {
      self.error(jqxhr.responseJSON.error);
    }
  });
}

/****** show ******/

var show = null;

main.Show = function(inst) {
  this.$this = $(this);
  this.name = inst.name;
  this.created = new Date(inst.created);
  this.updated = new Date(inst.updated);
  this.files = [];
  this.playlists = [];
};

main.Show.prototype.error = function(error) {
  this.last_error = error;
  this.$this.trigger('error');
};

main.Show.prototype.fetch_files = function() {
  var self = this;
  self.last_error = null;

  $.getJSON("/api/v1/shows/" + self.name + "/files", function(data) {
    self.files = [];
    for(var key in data.results) {
      self.files.push(new main.File(self, data.results[key]));
    }
    self.$this.trigger('update-files');
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      self.error("connection problem");
    } else {
      self.error(error);
    }
  });
};

main.Show.prototype.fetch_playlists = function() {
  var self = this;
  self.last_error = null;

  $.getJSON("/api/v1/shows/" + self.name + "/playlists", function(data) {
    self.playlists = [];
    for(var key in data.results) {
      self.playlists.push(new main.Playlist(self, data.results[key]));
    }
    self.$this.trigger('update-playlists');
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      self.error("connection problem");
    } else {
      self.error(error);
    }
  });
};

main.Show.prototype.add_file_upload = function(file) {
  var self = this;
  self.last_error = null;

  $.ajax({
    method: "POST",
    url: "/api/v1/shows/" + self.name + "/files",
    data: JSON.stringify({'source-uri': encodeURI('upload://' + file.name)}),
    success: function(data) {
      var f = new main.File(self, data);
      f.start_upload(file);
      self.fetch_files();
    },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      self.error("connection problem");
    } else {
      self.error(jqxhr.responseJSON.error);
    }
  });
}

main.Show.prototype.add_file_uri = function(file_uri) {
  var self = this;
  self.last_error = null;

  $.ajax({
    method: "POST",
    url: "/api/v1/shows/" + self.name + "/files",
    data: JSON.stringify({'source-uri': file_uri}),
    success: function(data) { self.fetch_files() },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    if(error == "") {
      self.error("connection problem");
    } else {
      self.error(jqxhr.responseJSON.error);
    }
  });
}

/****** file ******/

main.File = function(show, inst) {
  this.$show = show;
  this.id = inst.id;
  this.created = new Date(inst.created);
  this.updated = new Date(inst.updated);
  this.source = inst.source;
  this.metadata = inst.metadata;
  this.size = inst.size;
  this.duration = inst.duration;
};

main.File.prototype.delete = function() {
  var self = this;

  $.ajax({
    method: "DELETE",
    url: "/api/v1/shows/" + this.$show.name + "/files/" + this.id,
    success: function(data) { self.$show.fetch_files() },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    alert(jqxhr.responseJSON.error + "\n" + JSON.stringify(jqxhr.responseJSON.details, null, 2));
  });
};

main.File.prototype.edit = function(metadata) {
  var self = this;

  $.ajax({
    method: "PATCH",
    url: "/api/v1/shows/" + this.$show.name + "/files/" + this.id,
    data: JSON.stringify(metadata),
    success: function(data) { self.$show.fetch_files() },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    alert(jqxhr.responseJSON.error);
  });
}

main.File.prototype.start_upload = function(file) {
  var self = this;

  $.ajax({
    method: "GET",
    url: "/api/v1/shows/" + this.$show.name + "/files/" + this.id + "/import",
    data: {'wait-for': 'running'},
    success: function(data) { self.upload(file) },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    alert(jqxhr.responseJSON.error);
  });
}

main.File.prototype.upload = function(file) {
  var self = this;

  var flow = new Flow({
    target:"/api/v1/shows/" + this.$show.name + "/files/" + this.id + "/upload",
    chunkSize: 100 * 1024,
    prioritizeFirstAndLastChunk: true,
  });
  if(!flow.support) {
    return self.upload_simple(file);
  }
  flow.on('fileSuccess', function(file, message) {
    console.log(file, message);
  });
  flow.on('fileError', function(file, message) {
    console.log(file, message);
  });
  flow.addFile(file);
  flow.upload();
}

main.File.prototype.upload_simple = function(file) {
  var self = this;

  $.ajax({
    method: "PUT",
    url: "/api/v1/shows/" + this.$show.name + "/files/" + this.id + "/upload",
    contentType: "application/octet-stream",
    data: file,
    processData: false,
    success: function(data) { console.log(data) },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    alert(jqxhr.responseJSON.error);
  });
}


/****** playlist ******/

main.Playlist = function(show, inst) {
  this.$show = show;
  this.id = inst.id;
  this.created = new Date(inst.created);
  this.updated = new Date(inst.updated);
  this.entries = inst.entries;
};

main.Playlist.prototype.delete = function() {
  var self = this;

  $.ajax({
    method: "DELETE",
    url: "/api/v1/shows/" + this.$show.name + "/playlists/" + this.id,
    success: function(data) { self.$show.fetch_playlists() },
    dataType: "json"
  }).fail(function(jqxhr, textStatus, error) {
    alert(jqxhr.responseJSON.error + "\n" + JSON.stringify(jqxhr.responseJSON.details, null, 2));
  });
};

/***************** views *****************/

/****** authInfo ******/

var authInfoView = null;

main.AuthInfoView = function(model) {
  this.model = model;

  var self = this;
  $(this.model).on('update', function() {
    self.render_info();
  });
};

main.AuthInfoView.prototype.render_info = function() {
  var status = $('#authStatus');
  var btn = $('#btnLogInOut');
  var self = this;
  if (this.model.logged_in)  {
    status.html("you are logged in as <b>" + this.model.session.username + "</b>");
    btn.text("Logout")
    btn.removeClass('btn-primary').addClass('btn-danger');
    btn.off('click').on('click', function() {
      self.model.logout();
    });
  } else {
    status.html("you are <b>not</b> logged in");
    btn.text("Login")
    btn.removeClass('btn-danger').addClass('btn-primary');
    btn.off('click').on('click', function() {
      self.model.login();
    });
  }
};

/****** showList ******/

var showListView = null;

main.ShowListView = function(model) {
  this.model = model;

  var self = this;
  $(this.model).on('update', function() {
    self.render_list();
  });
  $(this.model).on('error', function() {
    self.render_error();
  });
};

main.ShowListView.prototype.render_list = function() {
  $('#error').hide();

  var list = $('#showSelect');
  $('option', list).remove();
  list.append($('<option>', {value:'__none__', text:'choose...'}))

  for (var name in this.model.shows) {
    list.append($('<option>', {value:name, text:name}));
  }
};

main.ShowListView.prototype.render_error = function() {
  $('#error-message').text('Error fetching show listing: ' + this.model.last_error)
  $('#error').show();
};

/****** show ******/

var showView = null;

main.ShowView = function(model) {
  this.model = model;
  this.$el = $('#hiddenTemplates .showTemplate').clone().removeClass('showTemplate');

  var self = this;
  var selectFiles = this.$el.find('a.showSelectFiles');
  var selectPlaylists = this.$el.find('a.showSelectPlaylists');
  var cardBody = self.$el.find('div.card-body');

  selectFiles.on('click', function() {
    selectFiles.addClass('active');
    selectPlaylists.removeClass('active');
    self.model.fetch_files();
    cardBody.text("... loading files ...");
  });
  selectPlaylists.on('click', function() {
    selectPlaylists.addClass('active');
    selectFiles.removeClass('active');
    self.model.fetch_playlists();
    cardBody.text("... loading playlists ...")
  });

  $(this.model).on('update-files', function() {
    self.render_files(cardBody);
  });
  $(this.model).on('update-playlists', function() {
    self.render_playlists(cardBody);
  });
  $(this.model).on('error', function() {
    self.render_error();
  });

  $('#showView').html(this.$el);
};

main.ShowView.prototype.render_files = function(body) {
  $('#error').hide();

  var fileListView = new main.FileListView(this.model.files, this);
	fileListView.render();
  body.html(fileListView.$el);
};

main.ShowView.prototype.render_playlists = function(body) {
  $('#error').hide();
  var fileListView = new main.PlaylistListView(this.model.playlists, this);
	fileListView.render();
  body.html(fileListView.$el);
};

main.ShowView.prototype.render_error = function() {
  $('#error-message').text(this.model.last_error)
  $('#error').show();
};

/****** files ******/

main.FileListView = function(model) {
  this.model = model;
  this.$el = $('#hiddenTemplates .filesTemplate').clone().removeClass('filesTemplate');
};

main.FileListView.prototype.render = function() {
  var table = this.$el.find("tbody");
  for (var i = 0; i < this.model.length; i++) {
    var fileView = new main.FileView(this.model[i], this);
    fileView.render();
    table.append(fileView.$el);
  }
};

main.FileView = function(model) {
  this.model = model;
  this.$el = $('#hiddenTemplates .fileTemplate tr').clone();
};

main.FileView.prototype.render = function() {
  this.$el.find('.fileId').text(this.model.id)
  this.$el.find('.fileArtist').text(this.model.metadata.artist)
  this.$el.find('.fileAlbum').text(this.model.metadata.album)
  this.$el.find('.fileTitle').text(this.model.metadata.title)
  this.$el.find('.fileDuration').text(format_duration(this.model.duration))

  var self = this;
  this.$el.find('.fileEdit').on('click', function() {
    self.render_edit_modal();
  });
  this.$el.find('.fileDelete').on('click', function() {
    self.model.delete();
  });
};

main.FileView.prototype.render_edit_modal = function() {
  var modal = $('#fileEditModal');
  var fileEditBtn = modal.find('button.fileEdit');

  modal.find('#fileEditID').text(this.model.$show.name + '/' + this.model.id);
  var fileEditArtist =  modal.find('#fileEditArtist').val(this.model.metadata.artist);
  var fileEditTitle = modal.find('#fileEditTitle').val(this.model.metadata.title);
  var fileEditAlbum = modal.find('#fileEditAlbum').val(this.model.metadata.album);

  var self = this;
  fileEditBtn.off('click').on('click', function() {
    var metadata = {};
    metadata.artist = fileEditArtist.val();
    metadata.title = fileEditTitle.val();
    metadata.album = fileEditAlbum.val();
    self.model.edit(metadata);
  });

  modal.find('#fileEditForm').off('submit').on('submit', function() {
    fileEditBtn.trigger('click');
    return false;
  });

  modal.on('shown.bs.modal', function(event) {
    fileEditArtist.focus();
  });

  modal.modal('show');
};

/****** playlists ******/

main.PlaylistListView = function(model) {
  this.model = model;
  this.$el = $('#hiddenTemplates .playlistsTemplate').clone().removeClass('playlistsTemplate');
};

main.PlaylistListView.prototype.render = function() {
  var table = this.$el.find("tbody");
  for (var i = 0; i < this.model.length; i++) {
    var playlistView = new main.PlaylistView(this.model[i], this);
    playlistView.render();
    table.append(playlistView.$el);
  }
};

main.PlaylistView = function(model) {
  this.model = model;
  this.$el = $('#hiddenTemplates .playlistTemplate tr').clone();
};

main.PlaylistView.prototype.render = function() {
  this.$el.find('.playlistId').text(this.model.id)
  this.$el.find('.playlistLength').text(this.model.entries.length)
  this.$el.find('.playlistCreated').text(format_datetime(this.model.created))

  var self = this;
  // TODO: implement playlist edit
  this.$el.find('.playlistDelete').on('click', function() {
    self.model.delete();
  });
};

/***************** controller *****************/

function modal_show_new_init() {
  var modal = $('#showNewModal');
  var showNewName = modal.find('#showNewName');
  var showNewBtn = modal.find('button.showNew');

  showNewName.on('keyup', function() {
    if(showNewName.val().length == 0) {
      showNewBtn.prop('disabled', true)
    } else {
      showNewBtn.prop('disabled', false)
    }
  });

  modal.find('#showNewForm').on('submit', function() {
    showNewBtn.trigger('click');
    return false;
  });

  modal.on('show.bs.modal', function(event) {
    showNewName.val("");

    showNewBtn.prop('disabled', true)
    showNewBtn.off('click').on('click', function() {
      shows.new_show(showNewName.val());
    });
  });

  modal.on('shown.bs.modal', function(event) {
    showNewName.focus();
  });
}

function modal_file_add_init() {
  var modal = $('#fileAddModal');
  // var fileAddURI = modal.find('#fileAddURI');
  var fileAddSelect = modal.find('#fileAddSelect');
  var fileAddBtn = modal.find('button.fileAdd');

  // fileAddURI.on('keyup', function() {
  //   if(fileAddURI.val().length == 0) {
  //     fileAddBtn.prop('disabled', true)
  //   } else {
  //     fileAddBtn.prop('disabled', false)
  //   }
  // });

  // modal.find('#fileAddForm').on('submit', function() {
  //   fileAddBtn.trigger('click');
  //   return false;
  // });

  modal.on('show.bs.modal', function(event) {
    // fileAddURI.val("");
    fileAddBtn.prop('disabled', true)
  });

  modal.on('shown.bs.modal', function(event) {
    // fileAddURI.focus();
    fileAddSelect.on('change', function() {
      var $this = this;
      fileAddBtn.prop('disabled', false)
      fileAddBtn.off('click').on('click', function() {
        show.add_file_upload($this.files[0]);
      });
    });
  });
}

function main_init() {
  auth = new main.AuthInfo();
  authInfoView = new main.AuthInfoView(auth);
  shows = new main.ShowList();
  showListView = new main.ShowListView(shows);

  modal_show_new_init();
  modal_file_add_init();

  auth.fetch();
  $('#showSelect').on('change', function(event) {
    var show_name = event.target.value;
    if(show_name != '__none__') {
      show = shows.shows[show_name];
      showView = new main.ShowView(show);
      show.fetch_files();
      $('button.showClone').prop('disabled', false);
      $('button.showAddFile').removeClass('disabled').attr('data-toggle', 'modal');
      $('button.showAddPlaylist').removeClass('disabled').attr('data-toggle', 'modal');
    } else {
      $('#showView').empty();
      $('button.showClone').prop('disabled', true);
      $('button.showAddFile').addClass('disabled').removeAttr('data-toggle');
      $('button.showAddPlaylist').addClass('disabled').removeAttr('data-toggle');
    }
  });
}

/*************************************************************************/
